import { Color, Point, Rect, Size } from "./engine/datatypes.ts";
import { D } from "./engine/public.ts";

export const FLOOR_CELL_SIZE: Size = new Size(48, 48);
export const CEILING_CELL_SIZE: Size = new Size(56, 56);
export const HEIGHT_IN_FEET = 12;
export const CENTER = new Point(192, 192);
export const MOULDING_SZ = new Size(1, 1);

export class GridArt {
  #at: Point;

  #floorCenter: Point;
  #ceilingCenter: Point;
  #floorTl: Point;
  #ceilingTl: Point;
  #floorBr: Point;
  #ceilingBr: Point;

  constructor(at: Point) {
    this.#at = at;
    this.#floorCenter = at.scale(FLOOR_CELL_SIZE).offset(CENTER);
    this.#ceilingCenter = at.scale(CEILING_CELL_SIZE).offset(CENTER);

    this.#floorTl = at
      .offset(new Point(-0.5, -0.5))
      .scale(FLOOR_CELL_SIZE)
      .offset(CENTER);
    this.#ceilingTl = at
      .offset(new Point(-0.5, -0.5))
      .scale(CEILING_CELL_SIZE)
      .offset(CENTER);
    this.#floorBr = at
      .offset(new Point(0.5, 0.5))
      .scale(FLOOR_CELL_SIZE)
      .offset(CENTER);
    this.#ceilingBr = at
      .offset(new Point(0.5, 0.5))
      .scale(CEILING_CELL_SIZE)
      .offset(CENTER);
  }

  get floorRect(): Rect {
    return new Rect(this.#floorTl, this.#floorBr.subtract(this.#floorTl));
  }

  drawFloor(color: Color) {
    D.fillRect(this.#floorTl, this.#floorBr.subtract(this.#floorTl), color);
  }

  #drawWallTop(color: Color) {
    let diff = Math.abs(this.#ceilingTl.y - this.#floorTl.y);
    let sign = Math.sign(this.#ceilingTl.y - this.#floorTl.y);
    // console.log(`diff, sign: ${diff}, ${sign}`)
    for (let dy = 0; dy <= diff; dy += 0.25) {
      // 0.25: fudge factor because we get two different lines
      let progress = dy / diff;
      let x0 = Math.floor(lerp(progress, this.#floorTl.x, this.#ceilingTl.x));
      let x1 = Math.ceil(lerp(progress, this.#floorBr.x, this.#ceilingBr.x));
      let y = this.#floorTl.y + sign * dy;

      if (dy == 0 || dy == diff) {
        // console.log(x0, x1, y);
      }
      D.fillRect(new Point(x0, y - 1), new Size(x1 - x0, 1), color);
    }
  }

  #drawWallLeft(color: Color) {
    let diff = Math.abs(this.#ceilingTl.x - this.#floorTl.x);
    let sign = Math.sign(this.#ceilingTl.x - this.#floorTl.x);
    // console.log(`diff, sign: ${diff}, ${sign}`)
    for (let dx = 0; dx <= diff; dx += 0.25) {
      // fudge factor because we get two different lines
      let progress = dx / diff;
      let y0 = Math.floor(lerp(progress, this.#floorTl.y, this.#ceilingTl.y));
      let y1 = Math.ceil(lerp(progress, this.#floorBr.y, this.#ceilingBr.y));

      let x = this.#floorTl.x + sign * dx;

      D.fillRect(new Point(x, y0), new Size(1, y1 - y0), color);
    }
  }

  drawWallTop(color: Color) {
    if (this.#at.y > 0) {
      return;
    }
    this.#drawWallTop(color);
  }

  drawWallLeft(color: Color) {
    if (this.#at.x > 0) {
      return;
    }
    this.#drawWallLeft(color);
  }

  drawWallBottom(color: Color) {
    if (this.#at.y < 0) {
      return;
    }
    new GridArt(this.#at.offset(new Point(0, 1))).#drawWallTop(color);
  }

  drawWallRight(color: Color) {
    if (this.#at.x < 0) {
      return;
    }
    new GridArt(this.#at.offset(new Point(1, 0))).#drawWallLeft(color);
  }

  drawMouldingTop(color: Color) {
    let lhs = this.#ceilingTl.offset(new Point(0, -MOULDING_SZ.h));
    D.fillRect(lhs, new Size(CEILING_CELL_SIZE.w, MOULDING_SZ.h), color);
  }

  drawMouldingTopLeft(color: Color) {
    D.fillRect(
      this.#ceilingTl.offset(new Point(-MOULDING_SZ.w, -MOULDING_SZ.h)),
      MOULDING_SZ,
      color,
    );
  }

  drawMouldingLeft(color: Color) {
    let lhs = this.#ceilingTl.offset(new Point(-MOULDING_SZ.w, 0));
    D.fillRect(lhs, new Size(MOULDING_SZ.w, CEILING_CELL_SIZE.h), color);
  }

  drawMouldingTopRight(color: Color) {
    D.fillRect(
      this.#ceilingTl.offset(new Point(CEILING_CELL_SIZE.w, -MOULDING_SZ.h)),
      MOULDING_SZ,
      color,
    );
  }

  drawMouldingBottom(color: Color) {
    let lhs = this.#ceilingTl.offset(new Point(0, CEILING_CELL_SIZE.h));
    D.fillRect(lhs, new Size(CEILING_CELL_SIZE.w, MOULDING_SZ.h), color);
  }

  drawMouldingBottomLeft(color: Color) {
    D.fillRect(
      this.#ceilingTl.offset(new Point(-MOULDING_SZ.w, CEILING_CELL_SIZE.h)),
      MOULDING_SZ,
      color,
    );
  }

  drawMouldingRight(color: Color) {
    let lhs = this.#ceilingTl.offset(new Point(CEILING_CELL_SIZE.w, 0));
    D.fillRect(lhs, new Size(MOULDING_SZ.w, CEILING_CELL_SIZE.h), color);
  }

  drawMouldingBottomRight(color: Color) {
    D.fillRect(
      this.#ceilingTl.offset(
        new Point(CEILING_CELL_SIZE.w, CEILING_CELL_SIZE.h),
      ),
      MOULDING_SZ,
      color,
    );
  }

  drawCeiling(color: Color) {
    D.fillRect(
      this.#ceilingTl,
      this.#ceilingBr.subtract(this.#ceilingTl),
      color,
    );
    // D.drawRect(this.#ceilingTl, this.#ceilingBr.subtract(this.#ceilingTl), FG_BOLD);
  }

  project(z: number): Point {
    let z2 = z / HEIGHT_IN_FEET;
    return new Point(
      lerp(z2, this.#floorCenter.x, this.#ceilingCenter.x),
      lerp(z2, this.#floorCenter.y, this.#ceilingCenter.y),
    );
  }
}

let lerp = (amt: number, x: number, y: number) => {
  if (amt <= 0) {
    return x;
  }
  if (amt >= 1) {
    return y;
  }
  return x + (y - x) * amt;
};
